'use strict';

const _                   = require('lodash');
const ResourceBaseManager = require('./kube-native/resource.base.controller');
const controllers         = require('./kube.controllers');
const config              = require('./../config');
const EventEmitter        = require('eventemitter2').EventEmitter2;
const rp                   = require('request-promise');
const kubeConfigBuilder = require('kube-config-builder');
const { Client } = require('kubernetes-client');
const kubeEndpoints = require('./kube-endpoints');

const SEMVER_REGEXP = /([0-9]+)\.([0-9]+)\.([0-9]+)(?:-([0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?(?:\+[0-9A-Za-z-]+)?/; // valid examples: 1.22.3, 2.1.0
const SEMVER_REGEXP_MINOR_GROUP = 2;

class KubeManager extends EventEmitter {


    static constructNewManagerPromise(credentials, requestOptiton = {}){
        return kubeConfigBuilder.Builder.ConstructKubeAPIConfigJson(credentials)
            .then(config => new KubeManager(config, requestOptiton))
            .then((kubeManager) => kubeManager.init());
    }

    constructor(clusterConfig, requestOptiton = {}) {
        super();
        this._clusterConfig = clusterConfig;
        this._requestOptiton = requestOptiton;
    }

    async init() {
        await this.buildControllers();
        return this;
    }

    async buildControllers() {
        const client = new Client({
            config: this._getClientConfig()
        });
        await client.loadSpec();
        const endpoints = await this._resolveEndpointsVersion(client);
        controllers(endpoints).map((controller) => {
            // create function like getDeplymentController for all controllers in /kube.controllers
            const name = _.camelCase(`get ${controller.name} controller`);
            this[name] = (ns) => {
                const opt = {
                    kubeClient: _.get(controller, 'kubeEndpoint.namespacePath') ? _.get(client, controller.kubeEndpoint.namespacePath)(ns) : client,
                    kubeManager: this,
                    selector: this._clusterConfig.name,
                    events: controller.events,
                    path: _.get(controller, 'kubeEndpoint.path'),
                    Entity: controller.Entity,
                    builders: controller.builders,
                    StatusAnalyzerClass: controller.StatusAnalyzer,
                    ns
                };

                let ctrl;
                if (controller.Controller) {
                    ctrl = new controller.Controller(opt);
                } else {
                    ctrl = new ResourceBaseManager(opt);
                }
                ctrl.onAny((event, data) => {
                    this.emit(event, _.merge(data, {
                        controller: controller.name
                    }));
                });
                return ctrl;
            };
        });
    }

    getName() {
        return _.get(this, '_clusterConfig.name');
    }

    _getClientConfig() {
        const timeout = parseInt(_.get(this, '_requestOptiton.timeout', '120000'));
        const request = {
            headers: {
                'User-Agent': config.userAgent
            },
            timeout,
        };
        return _.merge(this._clusterConfig, {
            request: {
                request,
            },
        });
    }

    async _resolveEndpointsVersion(client) {
        const { body } = await client.version.get();
        const matches = body.gitVersion.match(SEMVER_REGEXP);

        if (!matches) {
            return kubeEndpoints.v15;
        }
        const minor = Number(matches[SEMVER_REGEXP_MINOR_GROUP]);
        if (minor > 15) {
            return kubeEndpoints.v16;
        }
        return kubeEndpoints.v15;
    }

    /**
     * We using this verification for custom providers
     * @private
     */
    _verifyTokenAuth() {
        const options = {
            uri: this._clusterConfig.url + '/api/v1/',
            headers: {
                'Authorization' : `bearer ${_.get(this._clusterConfig, 'auth.bearer')}`
            },
            agentOptions: {
                ca: this._clusterConfig.ca
            }
        };

        return rp(options);
    }

    verifyCredentials() {

        return this._verifyTokenAuth()
            .catch(() => {
                // try auth as before
                return this.getNamespaceController().get();
            });
    }
}

module.exports = KubeManager;
